/*++
/* NAME
/*	log_adhoc 3
/* SUMMARY
/*	ad-hoc delivery event logging
/* SYNOPSIS
/*	#include <log_adhoc.h>
/*
/*	void	log_adhoc(id, stats, recipient, relay, dsn, status)
/*	const char *id;
/*	MSG_STATS *stats;
/*	RECIPIENT *recipient;
/*	const char *relay;
/*	DSN *dsn;
/*	const char *status;
/* DESCRIPTION
/*	This module logs delivery events in an ad-hoc manner.
/*
/*	log_adhoc() appends a record to the mail logfile
/*
/*	Arguments:
/* .IP queue
/*	The message queue name of the original message file.
/* .IP id
/*	The queue id of the original message file.
/* .IP stats
/*	Time stamps from different message delivery stages
/*	and session reuse count.
/* .IP recipient
/*	Recipient information. See recipient_list(3).
/* .IP sender
/*	The sender envelope address.
/* .IP relay
/*	Host we could (not) talk to.
/* .IP status
/*	bounced, deferred, sent, and so on.
/* .IP dsn
/*	Delivery status information. See dsn(3).
/* BUGS
/*	Should be replaced by routines with an attribute-value based
/*	interface instead of an interface that uses a rigid argument list.
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <string.h>

#ifdef STRCASECMP_IN_STRINGS_H
#include <strings.h>
#endif

/* Utility library. */

#include <msg.h>
#include <vstring.h>

/* Global library. */

#include <log_adhoc.h>

/* log_adhoc - ad-hoc logging */

void    log_adhoc(const char *id, MSG_STATS *stats, RECIPIENT *recipient,
		          const char *relay, DSN *dsn,
		          const char *status)
{
    static VSTRING *buf;
    int     delay;
    double  pdelay;			/* time before active queue */
    double  adelay;			/* active queue latency */
    double  sdelay;			/* connection set-up latency */
    double  xdelay;			/* transmission latency */
    struct timeval now;

    /*
     * Alas, we need an intermediate buffer for the pre-formatted result.
     * There are several optional fields, and we want to tweak some
     * formatting depending on delay values.
     */
    if (buf == 0)
	buf = vstring_alloc(100);

    /*
     * First, general information that identifies the transaction.
     */
    vstring_sprintf(buf, "%s: to=<%s>", id, recipient->address);
    if (recipient->orig_addr && *recipient->orig_addr
	&& strcasecmp(recipient->address, recipient->orig_addr) != 0)
	vstring_sprintf_append(buf, ", orig_to=<%s>", recipient->orig_addr);
    vstring_sprintf_append(buf, ", relay=%s", relay);

    /*
     * Next, performance statistics.
     * 
     * Use wall clock time to compute pdelay (before active queue latency) if
     * there is no time stamp for active queue entry. This happens when mail
     * is bounced by the cleanup server.
     * 
     * Otherwise, use wall clock time to compute adelay (active queue latency)
     * if there is no time stamp for hand-off to delivery agent. This happens
     * when mail was deferred or bounced by the queue manager.
     * 
     * Otherwise, use wall clock time to compute xdelay (message transfer
     * latency) if there is no time stamp for delivery completion. In the
     * case of multi-recipient deliveries the delivery agent may have to
     * specify the delivery completion time, so that all recipient records
     * have the same delay values.
     * 
     * Don't compute the sdelay (connection setup latency) if there is no time
     * stamp for connection setup completion.
     * 
     * XXX Instead of floating point, we can use integer math. This can also
     * simplify the formatting. Right now we have to work against %f and %g
     * to get the result that we want: no more than one digit of precision
     * for numbers < 10 and > 0.01.
     */
#define DELTA(x, y) \
	((x)->tv_sec - (y)->tv_sec + ((x)->tv_usec - (y)->tv_usec) / 1000000.0)

    if (stats->deliver_done.tv_sec)
	now = stats->deliver_done;
    else
	GETTIMEOFDAY(&now);
    delay = now.tv_sec - stats->incoming_arrival;
    pdelay = adelay = sdelay = xdelay = 0;
    if (stats->active_arrival.tv_sec) {
	pdelay = stats->active_arrival.tv_sec - stats->incoming_arrival;
	if (stats->agent_handoff.tv_sec) {
	    adelay = DELTA(&stats->agent_handoff, &stats->active_arrival);
	    if (stats->conn_setup_done.tv_sec) {
		sdelay = DELTA(&stats->conn_setup_done, &stats->agent_handoff);
		xdelay = DELTA(&now, &stats->conn_setup_done);
	    } else {
		xdelay = DELTA(&now, &stats->agent_handoff);
	    }
	} else {
	    adelay = DELTA(&now, &stats->active_arrival);
	}
    }
    if (stats->reuse_count > 0)
	vstring_sprintf_append(buf, ", conn_use=%d", stats->reuse_count + 1);

#define PRETTY_FORMATS(x) ((x) >= 9.5 || (x) < 0.095 ? "%.0f/" : "%.1g/")
#define PRETTY_FORMAT(x) ((x) >= 9.5 || (x) < 0.095 ? "%.0f" : "%.1g")

    vstring_sprintf_append(buf, ", delay=%d, delays=", delay);
    vstring_sprintf_append(buf, PRETTY_FORMATS(pdelay), pdelay);
    vstring_sprintf_append(buf, PRETTY_FORMATS(adelay), adelay);
    vstring_sprintf_append(buf, PRETTY_FORMATS(sdelay), sdelay);
    vstring_sprintf_append(buf, PRETTY_FORMAT(xdelay), xdelay);

    /*
     * Finally, the delivery status.
     */
    vstring_sprintf_append(buf, ", dsn=%s, status=%s (%s)",
			   dsn->status, status, dsn->reason);

    /*
     * Ship it off.
     */
    msg_info("%s", vstring_str(buf));
}
